import 'package:flutter/foundation.dart';
import 'package:flutter/gestures.dart';

import '../renderer/actor.dart';
import '../renderer/stage.dart';
import '../extensions/extensions.dart';
import 'events.dart';

mixin HasDraggablesMixin on Stage {
  @mustCallSuper
  void onDragStart(int pointerId, DragStartInfo info) {
    propagateToChildren<DraggableMixin>(
      (c) => c.handleDragStart(pointerId, info),
    );
  }

  @mustCallSuper
  void onDragUpdate(int pointerId, DragUpdateInfo info) {
    propagateToChildren<DraggableMixin>(
      (c) => c.handleDragUpdated(pointerId, info),
    );
  }

  @mustCallSuper
  void onDragEnd(int pointerId, DragEndInfo info) {
    propagateToChildren<DraggableMixin>(
      (c) => c.handleDragEnded(pointerId, info),
    );
  }

  @mustCallSuper
  void onDragCancel(int pointerId) {
    propagateToChildren<DraggableMixin>(
      (c) => c.handleDragCanceled(pointerId),
    );
  }
}

mixin DraggableMixin on Actor {
  bool _isDragged = false;
  bool get isDragged => _isDragged;

  bool onDragStart(DragStartInfo info) {
    return true;
  }

  bool onDragUpdate(DragUpdateInfo info) {
    return true;
  }

  bool onDragEnd(DragEndInfo info) {
    return true;
  }

  bool onDragCancel() {
    return true;
  }

  final List<int> _currentPointerIds = [];
  bool _checkPointerId(int pointerId) => _currentPointerIds.contains(pointerId);

  bool handleDragStart(int pointerId, DragStartInfo info) {
    if (containsPoint(info.position)) {
      _isDragged = true;
      _currentPointerIds.add(pointerId);
      return onDragStart(info);
    }
    return true;
  }

  bool handleDragUpdated(int pointerId, DragUpdateInfo info) {
    if (_checkPointerId(pointerId)) {
      return onDragUpdate(info);
    }
    return true;
  }

  bool handleDragEnded(int pointerId, DragEndInfo info) {
    if (_checkPointerId(pointerId)) {
      _isDragged = false;
      _currentPointerIds.remove(pointerId);
      return onDragEnd(info);
    }
    return true;
  }

  bool handleDragCanceled(int pointerId) {
    if (_checkPointerId(pointerId)) {
      _isDragged = false;
      _currentPointerIds.remove(pointerId);
      return onDragCancel();
    }
    return true;
  }
}

class DragStartInfo extends PositionInfo<DragStartDetails> with HandledField {
  DragStartInfo.fromDetails(
    Stage stage,
    DragStartDetails raw,
  ) : super(stage, raw.globalPosition, raw);
}

class DragUpdateInfo extends PositionInfo<DragUpdateDetails> with HandledField {
  final Vector2 delta;

  DragUpdateInfo.fromDetails(
    Stage stage,
    DragUpdateDetails raw,
  )   : delta = raw.delta.toVector2(),
        super(stage, raw.globalPosition, raw);
}

class DragEndInfo extends BaseInfo<DragEndDetails> with HandledField {
  DragEndInfo.fromDetails(
    Stage stage,
    DragEndDetails raw,
  ) : super(raw);
}
