import 'package:flutter/painting.dart';

import '../extensions/extensions.dart';

import 'anchor.dart';
import 'cache_renderer_mixin.dart';
import 'renderer.dart';
import 'matrix_transformer.dart';

class Actor extends Renderer with CacheRendererMixin {
  final _transform = MatrixTransform();
  Anchor _anchor;
  Vector2 _size;

  Actor({
    double? angle,
    Vector2? position,
    Vector2? skew,
    Vector2? size,
    Anchor? anchor,
    Vector2? scale,
    double? priority,
  })  : _anchor = anchor ?? Anchor.center,
        _size = size ?? Vector2.zero(),
        super(priority: priority) {
    _transform
      ..angle = angle ?? 0
      ..position = position ?? Vector2.zero()
      ..skew = skew ?? Vector2.zero()
      ..scale = scale ?? Vector2.all(1);

    _onModifiedSizeOrAnchor();
  }

  void setFrom(Actor other) {
    anchor = other.anchor;
    size = other.size;

    angle = other.angle;
    scale = other.scale;
    skew = other.skew;
    position = other.position;
  }

  double get angle => _transform.angle;
  set angle(double next) {
    if (_transform.angle != next) {
      refresh();
      onAngleChanged(_transform.angle, _transform.angle = next);
    }
  }

  void onAngleChanged(double prev, double next) {}

  Vector2 get position => _transform.position;
  set position(Vector2 next) {
    if (_transform.position != next) {
      refresh();
      onPositionChanged(_transform.position, _transform.position = next);
    }
  }

  void onPositionChanged(Vector2 prev, Vector2 next) {}

  Vector2 get scale => _transform.scale;
  set scale(Vector2 next) {
    if (_transform.scale != next) {
      refresh();
      onScaleChanged(_transform.scale, _transform.scale = next);
    }
  }

  void onScaleChanged(Vector2 prev, Vector2 next) {}

  Vector2 get skew => _transform.skew;
  set skew(Vector2 next) {
    if (_transform.skew != next) {
      refresh();
      onSkewChanged(_transform.skew, _transform.skew = next);
    }
  }

  void onSkewChanged(Vector2 prev, Vector2 next) {}

  Vector2 get size => _size;
  set size(Vector2 next) {
    if (_size != next) {
      _onModifiedSizeOrAnchor();
      refresh();
      onSizeChanged(_size, _size = next.clone());
    }
  }

  void onSizeChanged(Vector2 prev, Vector2 next) {}

  Anchor get anchor => _anchor;
  set anchor(Anchor next) {
    if (_anchor != next) {
      _onModifiedSizeOrAnchor();
      refresh();
      onAnchorChanged(_anchor, _anchor = next);
    }
  }

  void onAnchorChanged(Anchor prev, Anchor next) {}

  @override
  void transformCanvas(Canvas canvas, VoidCallback callback) {
    canvas
      ..save()
      ..transform(_transform.matrix.storage);
    callback();
    canvas.restore();
  }

  void _onModifiedSizeOrAnchor() {
    _transform.setOffsetBySizeAndAnchor(_size, _anchor);
  }

  @override
  bool containsPoint(Vector2 point) {
    final local = absoluteToLocal(point);
    return (local.x >= 0) && (local.y >= 0) && (local.x < size.x) && (local.y < size.y);
  }

  Vector2 parentPositionOf(Vector2 point) => _transform.localToGlobal(point);

  Vector2 parentPositionOfAnchor(Anchor anchor) {
    return anchor == this.anchor
        ? position
        : parentPositionOf(Vector2(anchor.x * size.x, anchor.y * size.y));
  }

  Vector2 absolutePositionOf(Vector2 point) {
    Vector2 parentPoint = parentPositionOf(point);
    Renderer? ancestor = parent;

    while (ancestor != null) {
      if (ancestor is Actor) {
        parentPoint = ancestor.parentPositionOf(parentPoint);
      }
      ancestor = ancestor.parent;
    }
    return parentPoint;
  }

  Vector2 absolutePositionOfAnchor(Anchor anchor) {
    return absolutePositionOf(Vector2(anchor.x * size.x, anchor.y * size.y));
  }

  Vector2 toLocal(Vector2 point) => _transform.globalToLocal(point);

  Vector2 absoluteToLocal(Vector2 point) {
    Renderer? ancestor = parent;

    while (ancestor != null) {
      if (ancestor is Actor) {
        return toLocal(ancestor.absoluteToLocal(point));
      }
      ancestor = ancestor.parent;
    }
    return toLocal(point);
  }

  Rect toRect() => _transform.toRect(_size);
  Rect toAbsoluteRect() => _transform.toRect(_size, absolutePositionOf);

  double get absoluteAngle {
    return ancestors()
        .whereType<Actor>()
        .fold<double>(angle, (totalAngle, c) => totalAngle + c.angle);
  }

  Vector2 get absolutePosition => absolutePositionOfAnchor(anchor);
  Vector2 get center => parentPositionOfAnchor(Anchor.center);
  Vector2 get absoluteCenter => absolutePositionOfAnchor(Anchor.center);
}
