import 'package:flutter/painting.dart';

import '../extensions/extensions.dart';
import 'anchor.dart';

class MatrixTransform {
  final Matrix4 _matrix = Matrix4.identity();
  bool _matrixCached = false;

  double _angle;
  Vector2 _offset;
  Vector2 _skew;
  Vector2 _scale;
  Vector2 _position;

  MatrixTransform()
      : _angle = 0,
        _offset = Vector2.zero(),
        _scale = Vector2.all(1),
        _skew = Vector2.zero(),
        _position = Vector2.zero();

  double get angle => _angle;
  set angle(double next) {
    _angle = next;
    _matrixCached = false;
  }

  Vector2 get offset => _offset;
  set offset(Vector2 next) {
    _offset = next.clone();
    _matrixCached = false;
  }

  void setOffsetBySizeAndAnchor(Vector2 size, Anchor anchor) {
    offset = Vector2(-anchor.x * size.x, -anchor.y * size.y);
  }

  Vector2 get scale => _scale;
  set scale(Vector2 next) {
    _scale = next.clone();
    _matrixCached = false;
  }

  Vector2 get skew => _skew;
  set skew(Vector2 next) {
    _skew = next.clone();
    _matrixCached = false;
  }

  Vector2 get position => _position;
  set position(Vector2 next) {
    _position = next.clone();
    _matrixCached = false;
  }

  Vector2 globalToLocal(Vector2 point) {
    final transform = matrix.clone();
    final det = transform.invert(); // side effect

    if (det == 0.0) {
      return Vector2.zero();
    }

    final Vector3 n = Vector3(0.0, 0.0, 1.0);
    final Vector3 i = transform.perspectiveTransform(Vector3(0.0, 0.0, 0.0));
    final Vector3 d = transform.perspectiveTransform(Vector3(0.0, 0.0, 1.0)) - i;
    final Vector3 s = transform.perspectiveTransform(Vector3(point.x, point.y, 0.0));
    final Vector3 p = s - d * (n.dot(s) / n.dot(d));

    return Vector2(p.x, p.y);
  }

  Vector2 localToGlobal(Vector2 point) {
    return MatrixUtils.transformPoint(matrix, point.toOffset()).toVector2();
  }

  void flipHorizontally() {
    scale = Vector2(-_scale.x, _scale.y);
  }

  void flipVertically() {
    scale = Vector2(_scale.x, -_scale.y);
  }

  void flip() {
    scale = Vector2(-_scale.x, -_scale.y);
  }

  Matrix4 get matrix {
    if (_matrixCached) {
      return _matrix;
    }

    _matrix
      ..setIdentity()
      ..translate(_position.x, _position.y);

    if (_angle != 0) {
      _matrix.rotateZ(_angle);
    }

    if (_scale.x != 1 || _scale.y != 1) {
      _matrix.scale(_scale.x, _scale.y);
    }

    if (_offset.x != 0 || _offset.y != 0) {
      _matrix.translate(_offset.x, _offset.y);
    }

    if (_skew.x != 0 || _skew.y != 0) {
      _matrix.multiply(Matrix4.skew(_skew.x, _skew.y));
    }

    _matrixCached = true;
    return _matrix;
  }

  Rect toRect(Vector2 size, [Vector2 Function(Vector2 point)? projector]) {
    final convert = projector ?? localToGlobal;

    final topLeft = convert(Vector2.zero());
    final bottomRight = convert(size);

    if (_angle == 0.0) {
      return Rect.fromPoints(topLeft.toOffset(), bottomRight.toOffset());
    }

    final topRight = convert(Vector2(size.x, 0));
    final bottomLeft = convert(Vector2(0, size.y));
    final xs = [topLeft.x, topRight.x, bottomLeft.x, bottomRight.x]..sort();
    final ys = [topLeft.y, topRight.y, bottomLeft.y, bottomRight.y]..sort();

    return Rect.fromLTRB(xs.first, ys.first, xs.last, ys.last);
  }
}
