import '../gestures/drag_detector.dart';
import '../gestures/hover_detector.dart';
import '../gestures/scroll_detector.dart';
import 'package:flutter/gestures.dart';
import 'package:flutter/widgets.dart';

import '../gestures/tap_detector.dart';
import '../renderer/stage.dart';
import '../extensions/extensions.dart';

const _perFrame = Duration(milliseconds: 16);

Widget applyMouseMoveGesturesDetectors(Stage stage, Widget child) {
  if (stage is HasHoverablesMixin) {
    Future? _future;
    PointerHoverEvent? _event;
    int _flag = 0;

    return MouseRegion(
      onEnter: (event) {
        stage.onMouseEnter(PointerEnterInfo.fromDetails(stage, event));
      },
      onExit: (event) {
        if (_future != null && _event != null) {
          stage.onMouseMove(PointerHoverInfo.fromDetails(stage, _event!));
        }

        _future = null;
        _event = null;
        ++_flag;

        stage.onMouseLeave(PointerLeaveInfo.fromDetails(stage, event));
      },
      onHover: (event) {
        _event = event;
        if (_future != null) return;

        final flag = _flag;
        _future = Future.delayed(_perFrame, () {
          if (flag != _flag) return;
          if (_event != null) {
            stage.onMouseMove(PointerHoverInfo.fromDetails(stage, _event!));
            _event = null;
          }
          _future = null;
        });
      },
      child: child,
    );
  }

  return child;
}

Widget applyScrollDetector(Stage stage, Widget child) {
  if (stage is ScrollDetectorMixin) {
    return Listener(
      onPointerSignal: (event) {
        if (event is PointerScrollEvent) {
          stage.onScroll(PointerScrollInfo.fromDetails(stage, event));
        }
      },
      child: child,
    );
  }

  return child;
}

Widget applyBasicGesturesDetectors(Stage stage, Widget child) {
  final isTapMixin = stage is TapDetectorMixin;
  final isDoubleTapMixin = stage is DoubleTapDetectorMixin;

  if (!isTapMixin && !isDoubleTapMixin) return child;

  final onlyDoubleTap = isDoubleTapMixin && !isTapMixin;

  int tapDownTime = 0;
  bool doubleTapMode = false;

  return GestureDetector(
    key: const ObjectKey('BasicGesturesDetector'),
    behavior: HitTestBehavior.opaque,

    // Taps & Double Taps
    onTap: isTapMixin
        ? () {
            isDoubleTapMixin && doubleTapMode
                ? (stage as DoubleTapDetectorMixin).onDoubleTap()
                : stage.onTap();
            tapDownTime = doubleTapMode ? 0 : tapDownTime;
            doubleTapMode = false;
          }
        : null,
    onTapDown: isTapMixin
        ? (TapDownDetails details) {
            final info = TapDownInfo.fromDetails(stage, details);
            final now = DateTime.now().microsecondsSinceEpoch;

            doubleTapMode = now - tapDownTime <= kDoubleTapTimeout.inMicroseconds;
            tapDownTime = now;
            isDoubleTapMixin && doubleTapMode
                ? (stage as DoubleTapDetectorMixin).onDoubleTapDown(info)
                : stage.onTapDown(info);
          }
        : null,
    onTapUp: isTapMixin
        ? (TapUpDetails details) {
            isDoubleTapMixin && doubleTapMode
                ? null
                : stage.onTapUp(TapUpInfo.fromDetails(stage, details));
          }
        : null,
    onTapCancel: isTapMixin
        ? () {
            isDoubleTapMixin && doubleTapMode
                ? (stage as DoubleTapDetectorMixin).onDoubleTapCancel()
                : stage.onTapCancel();
            tapDownTime = 0;
            doubleTapMode = false;
          }
        : null,

    // Double Taps
    onDoubleTapDown: onlyDoubleTap
        ? (TapDownDetails details) => stage.onDoubleTapDown(TapDownInfo.fromDetails(stage, details))
        : null,
    onDoubleTap: onlyDoubleTap ? () => stage.onDoubleTap() : null,
    onDoubleTapCancel: onlyDoubleTap ? () => stage.onDoubleTapCancel() : null,

    child: child,
  );
}

Widget applyAdvancedGesturesDetectors(Stage stage, Widget child) {
  final isDraggableMixin = stage is HasDraggablesMixin;

  if (!isDraggableMixin) return child;

  final Map<Type, GestureRecognizerFactory> gestures = {};
  int lastGeneratedDragId = 0;

  void addAndConfigureRecognizer<T extends GestureRecognizer>(
    T Function() ts,
    void Function(T) bindHandlers,
  ) {
    gestures[T] = GestureRecognizerFactoryWithHandlers<T>(ts, bindHandlers);
  }

  void addDragRecognizer(Stage stage, Drag Function(int, DragStartInfo) config) {
    addAndConfigureRecognizer(
      () => ImmediateMultiDragGestureRecognizer(),
      (ImmediateMultiDragGestureRecognizer instance) {
        instance.onStart = (Offset o) {
          final pointerId = lastGeneratedDragId++;

          final global = o;
          final local = stage.convertGlobalToLocalCoordinate(global.toVector2()).toOffset();
          final details = DragStartDetails(
            localPosition: local,
            globalPosition: global,
          );

          return config(pointerId, DragStartInfo.fromDetails(stage, details));
        };
      },
    );
  }

  if (isDraggableMixin) {
    addDragRecognizer(stage, (int pointerId, DragStartInfo position) {
      stage.onDragStart(pointerId, position);
      return _DragEvent(
        stage,
        onUpdate: (details) => stage.onDragUpdate(pointerId, details),
        onEnd: (details) => stage.onDragEnd(pointerId, details),
        onCancel: () => stage.onDragCancel(pointerId),
      );
    });
  }

  return RawGestureDetector(
    gestures: gestures,
    behavior: HitTestBehavior.opaque,
    child: child,
  );
}

class _DragEvent extends Drag {
  final Stage stage;
  final void Function(DragUpdateInfo) onUpdate;
  final VoidCallback onCancel;
  final void Function(DragEndInfo) onEnd;

  Future? _future;
  DragUpdateDetails? _details;

  _DragEvent(
    this.stage, {
    required this.onUpdate,
    required this.onEnd,
    required this.onCancel,
  });

  @override
  void update(DragUpdateDetails details) {
    _details = details;

    if (_future != null) return;

    _future = Future.delayed(_perFrame, () {
      if (_details != null) {
        onUpdate(DragUpdateInfo.fromDetails(stage, _details!));
        _details = null;
      }
      _future = null;
    });
  }

  @override
  void cancel() {
    _finishUpdate();
    onCancel();
  }

  @override
  void end(DragEndDetails details) {
    _finishUpdate();
    onEnd(DragEndInfo.fromDetails(stage, details));
  }

  void _finishUpdate() {
    if (_future != null && _details != null) {
      onUpdate(DragUpdateInfo.fromDetails(stage, _details!));
    }

    _future = null;
    _details = null;
  }
}
